//
//  AliCDNImageFlutterPlugin.m
//  Runner
//
//  Created by wqyfavor on 2020/1/3.
//  Copyright © 2020 Alibaba. All rights reserved.
//

#import "AlicdnImageFlutterPlugin.h"
#import <TBCDNImage/TBCDNImage.h>

@interface AlicdnImageOperation : NSObject <FlutterExternalAdapterImageRequest>

@property (nonatomic, assign) BOOL canceled;

/*
 CDN provides functionality to rewrite the URL by configs.
 But after parsing the URL may be invalid, so we must implement
 retry logic. In this case the downloading operation may change.
 So we must use this class to hold the current operation.
 */
@property (atomic, strong) id<SDWebImageOperation> operation;

@end

@implementation AlicdnImageOperation

- (instancetype)initWithOperation:(id<SDWebImageOperation>)operation
{
  if (self = [super init]) {
    _operation = operation;
  }
  return self;
}

- (void)cancel {
  _canceled = YES;
  [_operation cancel];
}

@end

@interface AlicdnImageProvider : NSObject <FlutterExternalAdapterImageProvider>
{
  NSLock* _requestIdLock;
  NSUInteger _requestId;
  
  NSLock* _requestsLock;
  NSMutableDictionary<NSNumber*, AlicdnImageOperation*>* _requests;
}
@end

@implementation AlicdnImageProvider

- (instancetype)init {
  if (self = [super init]) {
    _requestIdLock = [[NSLock alloc] init];
    
    _requestsLock = [[NSLock alloc] init];
    _requests = [[NSMutableDictionary alloc] init];
  }
  return self;
}

- (AlicdnImageOperation*)setRequest:(NSUInteger)rid operation:(id<SDWebImageOperation>)operation
{
  AlicdnImageOperation* existingOperation = nil;
  @try {
    [_requestsLock lock];
    existingOperation = _requests[@(rid)];
    if (existingOperation == nil) {
      existingOperation = [[AlicdnImageOperation alloc] initWithOperation:operation];
      _requests[@(rid)] = existingOperation;
    }
    else {
      existingOperation.operation = operation;
    }
    
    if ([_requests count] > 25) {
      // Clear canceled requests.
      NSMutableArray* canceledKeys = [[NSMutableArray alloc] init];
      for (NSNumber* key in [_requests allKeys]) {
        AlicdnImageOperation* op = _requests[key];
        if (op.canceled) {
          [canceledKeys addObject:key];
        }
      }
      [_requests removeObjectsForKeys:canceledKeys];
    }
  }
  @finally {
    [_requestsLock unlock];
  }
  return existingOperation;
}

- (void)removeRequest:(NSUInteger)rid
{
  @try {
    [_requestsLock lock];
    [_requests removeObjectForKey:@(rid)];
  }
  @finally {
    [_requestsLock unlock];
  }
}

- (id<FlutterExternalAdapterImageRequest>)request:(NSString*)url
                                      targetWidth:(NSInteger)targetWidth
                                     targetHeight:(NSInteger)targetHeight
                                       parameters:(NSDictionary<NSString*, NSString*>*)parameters
                                        extraInfo:(NSDictionary<NSString*, NSString*>*)extraInfo
                                         callback:(void(^)(UIImage* image))callback {
  //NSLog(@"[AliCDNImage] Requesting. Size(%ld, %ld), URL: %@", targetWidth, targetHeight, url);
  
  [_requestIdLock lock];
  NSUInteger rid = _requestId ++;
  [_requestIdLock unlock];
  
  NSURL* originalURL = [NSURL URLWithString:url];
  if (originalURL == nil) {
    callback(nil);
  }
  
  // targetWidth/Height is in points. CDN parser also request image frame size in points.
  CGSize imageViewSize = CGSizeZero;
  if (targetWidth > 0 && targetHeight > 0) {
    imageViewSize.width = targetWidth;
    imageViewSize.height = targetHeight;
  }
  
  ImageAdaptorConfig* cdnConfig = [[ImageAdaptorConfig alloc] init];
  cdnConfig.cdnImageQuality = CDN_IMAGE_QUALITY_CONFIG;
  cdnConfig.cdnImageWebp = CDN_IMAGE_WEBP_CONFIG;
  cdnConfig.cdnImageHeic = CDN_IMAGE_HEIC_CONFIG;
  cdnConfig.cdnImageSharp = CDN_IMAGE_SHARP_CONFIG;
  cdnConfig.cdnImageGif = CDN_IMAGE_GIF_PLAY;
  
  bool originalImage = false;
  
  if (parameters) {
    // sharp
    if ([parameters[@"sharp"] isEqualToString:@"Never"]) {
      cdnConfig.cdnImageSharp = CDN_IMAGE_SHARP_NONE;
    }
    else if ([parameters[@"sharp"] isEqualToString:@"Always"]) {
      cdnConfig.cdnImageSharp = CDN_IMAGE_SHARP_USE;
    }

    // quality
    NSString* qualityValue = parameters[@"quality"];
    if ([qualityValue isEqualToString:@"Original"]) {
      originalImage = true;
    }
    else if ([qualityValue isEqualToString:@"Low"]) {
      cdnConfig.cdnImageQuality = CDN_IMAGE_QUALITY_Q_30;
    }
    else if ([qualityValue isEqualToString:@"Normal"]) {
      cdnConfig.cdnImageQuality = CDN_IMAGE_QUALITY_Q_60;
    }
    else if ([qualityValue isEqualToString:@"High"]) {
      cdnConfig.cdnImageQuality = CDN_IMAGE_QUALITY_Q_90;
    }
    
    // webp
    if ([parameters[@"webp"] isEqualToString:@"Never"]) {
      cdnConfig.cdnImageWebp = CDN_IMAGE_WEBP_NONE;
    }
    else if ([parameters[@"webp"] isEqualToString:@"Always"]) {
      cdnConfig.cdnImageWebp = CDN_IMAGE_WEBP_USE;
    }
    
    // heic
    if ([parameters[@"heic"] isEqualToString:@"Never"]) {
      cdnConfig.cdnImageHeic = CDN_IMAGE_HEIC_NONE;
    }
    else if ([parameters[@"heic"] isEqualToString:@"Always"]) {
      cdnConfig.cdnImageHeic = CDN_IMAGE_HEIC_USE;
    }
    
    // multiframe, gif
    if ([parameters[@"multiframe"] isEqualToString:@"Compress"]) {
      cdnConfig.cdnImageGif = CDN_IMAGE_GIF_COMPRESS;
    }
  }
  
  NSString* module = @"AliFlutter";
  
  /*
   手淘图片库逻辑：
   链接经过parse修改后，可能存在下载失败的情况，而图片库[UIImage setImageWithURL]相关方法会进行404重试，download相关方法不会。
   所以对download方法需要自主进行重试。而且download方法内也会调用CDN parse逻辑，所以自己parse过后，调用download方法，就要强制
   传(imageOption | SDWebImageNoParse)，不让图片库再parse一次了。
   */
  
  // Let CDN to parse the url.
  NSURL* parsedURL = originalImage ? originalURL :
                            [TBCDNImageURLParser parseImageURLForCDNURL:originalURL
                                                                 module:module
                                                              imageSize:CGSizeZero
                                                               viewSize:imageViewSize
                                                                cutType:ImageCutType_None
                                                                 config:cdnConfig];
  if (parsedURL == nil) {
    parsedURL = originalURL;
  }
  
  BOOL urlChanged = ![parsedURL isEqual:originalURL];
  
  __weak AlicdnImageProvider* wself = self;
  __block BOOL alreadyFinished = NO;
  
  id<SDWebImageOperation> operation =
    [[SDWebImageManager sharedManager] downloadImageWithURL:parsedURL
                                                     module:module
                                                  imageSize:CGSizeZero
                                                   viewSize:CGSizeZero
                                                    cutType:ImageCutType_None
                                                    options:SDWebImageNoParse
                                                    extInfo:extraInfo
                                                   progress:nil
                                                  completed:^(UIImage *image,
                                                              NSError *error,
                                                              SDImageCacheType cacheType,
                                                              BOOL finished,
                                                              NSURL *imageURL) {
      alreadyFinished = YES;
      if (error != nil) {
        if (urlChanged) {
          // CDN changed the url, we should retry with the original url.
          id<SDWebImageOperation> retryOperation = [[SDWebImageManager sharedManager] downloadImageWithURL:originalURL
                                                                                                    module:module
                                                                                                 imageSize:CGSizeZero
                                                                                                  viewSize:CGSizeZero
                                                                                                   cutType:ImageCutType_None
                                                                                                   options:SDWebImageNoParse
                                                                                                   extInfo:extraInfo
                                                                                                  progress:nil
                                                                                                 completed:^(UIImage *image,
                                                                                                             NSError *error,
                                                                                                             SDImageCacheType cacheType,
                                                                                                             BOOL finished,
                                                                                                             NSURL *imageURL) {
            callback(image);
            [wself removeRequest:rid];
          }];
          [wself setRequest:rid operation:retryOperation]; // Update the operation.
        }
        else {
          callback(nil); // CDN parser did not rewrite the url, no need to retry.
          [wself removeRequest:rid];
        }
      }
      else {
        //NSLog(@"[AliCDNImage] Get Image. Size(%f, %f), URL: %@", image.size.width, image.size.height, url);
        
        callback(image);
        [wself removeRequest:rid];
      }
  }];
  
  if (alreadyFinished) {
    return nil;
  }
  else {
    return [self setRequest:rid operation:operation];
  }
}

- (void)log:(NSString*)log {
  // TODO implement TLog.
  NSLog(@"[AliCDNImageProvider] %@", log);
}

@end

@implementation AlicdnImageFlutterPlugin

+ (void)registerWithRegistrar:(NSObject<FlutterPluginRegistrar>*)registrar {
  // Do nothing. This is not a valid plugin.
}

#if 1
+ (void)load {
  [self install];
}
#endif

+ (void)install {
  [FlutterEngine installExternalAdapterImageProvider:[[AlicdnImageProvider alloc] init]];
}

@end
