import 'dart:async';
import 'package:flutter/widgets.dart';

// Format policies.
enum AliCDImageSharp {
  Default,
  Never,
  Always,
}

enum AliCDNImageQuality {
  Default,
  Original, // Request the original image source without CDN processing. (PNG & JPEG)
  Low, // q30, only works for JPEG.
  Normal, // q60, only works for JPEG.
  High, // q90, only works for JPEG.
}

enum AliCDNImageWebP {
  Default,
  Never, // Never downloading WebP format.
  Always,
}

enum AliCDNImageHEIC {
  Default,
  Never, // Never downloading HEIC format.
  Always,
}

// GIF/APNG
enum AliCDNImageMultiFrame {
  /* 
  Allow animation, but the image library may still return static one frame image
  if the decoded data of the GIF is too large.
   */
  Default,

  /*
  Downgrade the animated image to static one frame image.
  */
  Compress,
}

class AliCDNImagePolicy {
  static final AliCDNImagePolicy Default = const AliCDNImagePolicy();

  const AliCDNImagePolicy(
    {
      this.sharp = AliCDImageSharp.Default,
      this.quality = AliCDNImageQuality.Default,
      this.webp = AliCDNImageWebP.Default,
      this.heic = AliCDNImageHEIC.Default,
      this.multiframe = AliCDNImageMultiFrame.Default,
    }
  );

  final AliCDImageSharp sharp;
  final AliCDNImageQuality quality;
  final AliCDNImageWebP webp;
  final AliCDNImageHEIC heic;
  final AliCDNImageMultiFrame multiframe;

  @override
  bool operator ==(dynamic other) {
    if (other.runtimeType != runtimeType)
      return false;
    final AliCDNImagePolicy typedOther = other;
    return sharp == typedOther.sharp
      && quality == typedOther.quality
      && webp == typedOther.webp
      && heic == typedOther.heic
      && multiframe == typedOther.multiframe;
  }
}

Map<String, String> _AliCDNImagePolicy2Parameters(AliCDNImagePolicy policy) {
  if (policy == null) {
    return null;
  }

  Map<String, String> result = {};
  if (policy.sharp != AliCDImageSharp.Default) {
    result["sharp"] = policy.sharp.toString().split('.').last;
  }
  if (policy.quality != AliCDNImageQuality.Default) {
    result["quality"] = policy.quality.toString().split('.').last;
  }
  if (policy.webp != AliCDNImageWebP.Default) {
    result["webp"] = policy.webp.toString().split('.').last;
  }
  if (policy.heic != AliCDNImageHEIC.Default) {
    result["heic"] = policy.heic.toString().split('.').last;
  }
  if (policy.multiframe != AliCDNImageMultiFrame.Default) {
    result["multiframe"] = policy.multiframe.toString().split('.').last;
  }
  return result;  
}

class AliCDNImage extends Image {
  AliCDNImage(String url,
    {
      // Flutter Image widget default arguments.
      Key key,
      semanticLabel,
      excludeFromSemantics = false,
      width,
      height,
      color,
      colorBlendMode,
      fit,
      alignment = Alignment.center,
      repeat = ImageRepeat.noRepeat,
      centerSlice,
      matchTextDirection = false,
      gaplessPlayback = false,
      filterQuality = FilterQuality.low,

      /*
      Set placeholder provider. This provider could be AssetImage, 
      FileImage, NetworkImage or even another ExternalAdapterImage
      */
      ImageProvider placeholderProvider,
      
      // CDN parameters
      double targetWidth, // Image request width in point
      double targetHeight, // Image request height in point
      AliCDNImagePolicy policy,

      // Extra info passed to image library.
      Map<String, String> extraInfo,
   }): super.external_adapter(
     url,
     key: key,
     frameBuilder: frameBuilder,
     loadingBuilder: loadingBuilder,
     semanticLabel: semanticLabel,
     excludeFromSemantics: excludeFromSemantics,
     width: width ?? targetWidth,
     height: height ?? targetHeight,
     color: color,
     colorBlendMode: colorBlendMode,
     fit: fit,
     alignment: alignment,
     repeat: repeat,
     centerSlice: centerSlice,
     matchTextDirection: matchTextDirection,
     gaplessPlayback: gaplessPlayback,
     filterQuality: filterQuality,

     // CDN parameters
     targetWidth: (targetWidth != null) ? targetWidth.toInt() : null,
     targetHeight: (targetHeight != null) ? targetHeight.toInt() : null,
     parameters: _AliCDNImagePolicy2Parameters(policy),
     extraInfo: extraInfo,
     placeholderProvider: placeholderProvider,
   );
}
